Components.utils.import("resource://gre/modules/Services.jsm");

function startup(data,reason)
{
    debug("startup");
    forEachOpenWindow(addPageListener);
    Services.wm.addListener(windowListener);
}

function shutdown(data,reason)
{
    if (reason == APP_SHUTDOWN)
        return;  // Just exit; don't need to bother unregistering listeners

    debug("shutdown");
    Services.wm.removeListener(windowListener);
    forEachOpenWindow(unloadWindow);
}

function forEachOpenWindow(todo)
{
    var openWindows = Services.wm.getEnumerator("navigator:browser");
    while (openWindows.hasMoreElements())
        todo(openWindows.getNext());
}

function unloadWindow(window)  // The actual window unloading is done in listeners to the "aboutconfigdesc-unload" custom event
{
    var event = window.document.createEvent("Event");
    event.initEvent("aboutconfigdesc-unload",false,false);
    window.dispatchEvent(event);
}

var windowListener =
{
    onOpenWindow : function(xulWindow)
    {
        var window = xulWindow.QueryInterface(Components.interfaces.nsIInterfaceRequestor)
                              .getInterface(Components.interfaces.nsIDOMWindow);
        function onWindowLoad()
        {
            window.removeEventListener("load",onWindowLoad,false);
            if (window.document.documentElement.getAttribute("windowtype") == "navigator:browser")
                addPageListener(window);
        }
        window.addEventListener("load",onWindowLoad,false);
    },
    onCloseWindow : function(xulWindow)
    {
        var window = xulWindow.QueryInterface(Components.interfaces.nsIInterfaceRequestor)
                              .getInterface(Components.interfaces.nsIDOMWindow);
        if (window.document.documentElement.getAttribute("windowtype") == "navigator:browser")
            unloadWindow(window);
    },
    onWindowTitleChange : function() {}
};

function addPageListener(window)
{
    debug("addPageListener");

    function checkForAboutConfig()
    {
        if (window.content.document.location.href != "about:config")
            return;

        if (window.content.document.readyState != "complete")  // If the document isn't loaded yet, wait
        {
            function onDocLoad()
            {
                window.content.removeEventListener("load",onDocLoad,false);
                checkForAboutConfig();
            }
            window.content.addEventListener("load",onDocLoad,false);
            return;
        }

        loadInToAboutConfig(window);
    }

    var progressListener =
    {
        onLocationChange : function() { checkForAboutConfig(); },
        onProgressChange : function() {},
        onSecurityChange : function() {},
        onStateChange : function() {},
        onStatusChange : function() {}
    };
    window.getBrowser().addProgressListener(progressListener);  // Because of the need for this Thunderbird can't be supported

    checkForAboutConfig();

    function unload(event)
    {
        debug("unload window");

        window.removeEventListener("unload",unload,false);
        window.removeEventListener("aboutconfigdesc-unload",unload,false);

        window.getBrowser().removeProgressListener(progressListener);
    }
    window.addEventListener("unload",unload,false);
    window.addEventListener("aboutconfigdesc-unload",unload,false);
}

function loadInToAboutConfig(xulWindow)
{
    try
    {
        var window = xulWindow.content;

        if (window.document.getElementById("descCol"))  // If the new column is present then we've already loaded into this page
            return;

        if (!window.gPrefArray.length)  // If the prefs haven't loaded up yet, wait
        {
            window.setTimeout(function() {
                loadInToAboutConfig(xulWindow);
            }, 250);
            return;
        }

        debug("loadInToAboutConfig");

        // Load all of the source comments from the files
        var defaultPrefsComments = loadDefaultPrefs();

        // Add new description column to tree XUL to show the new data
        var tree = window.document.getElementById("configTree");
        var treecols = tree.firstChild;
        var newSplitter = window.document.createElement("splitter");
        newSplitter.setAttribute("class", "tree-splitter");
        var newColumn = window.document.createElement("treecol");
        newColumn.setAttribute("id", "descCol");
        newColumn.setAttribute("label", "Source Comment");  // TODO: L10N?
        newColumn.setAttribute("flex", 12);
        treecols.appendChild(newSplitter);
        treecols.appendChild(newColumn);

        // Add description data to the object storing all pref data for parsing into the tree view
        var prefs = window.gPrefArray;
        for (var i in prefs)
        {
            var prefName = prefs[i].prefCol;
            var prefDesc = defaultPrefsComments[prefName];
            prefs[i].descCol = prefDesc ? prefDesc : "" ;
        }

        // Add preference observer to put the description back in when a line is refreshed
        var prefBranch = Services.prefs.getBranch(null).QueryInterface(Components.interfaces.nsIPrefBranch2);  // TODO: drop QI if Gecko 10.0-12.0 support is dropped
        var prefListener =
        {
            observe: function(subject,topic,prefName)
            {
                if (topic != "nsPref:changed")
                    return;
                window.setTimeout(function()  // Need to update after the built-in listener does its work
                {
                    debug("refresh pref");
                    var prefObj = window.gPrefHash[prefName];
                    if (!prefObj)
                        return;
                    var viewIndex = window.getViewIndexOfPref(prefObj);
                    var arrayIndex = window.getIndexOfPref(prefObj);
                    var prefDesc = defaultPrefsComments[prefName];
                    window.gPrefArray[arrayIndex].descCol = prefDesc ? prefDesc : "" ;
                    if (viewIndex >= 0)
                    {
                        window.gPrefView[viewIndex] = window.gPrefArray[arrayIndex];
                        tree.treeBoxObject.invalidateRow(viewIndex);
                    }
                }, 100);
            }
        };
        prefBranch.addObserver("",prefListener,false);

        function descColSortFunction(x,y)  // FIXME: sorts to the bottom of the tree view (though, value sorting does this too, so I'm just copying the bug)
        {
            if (x.descCol > y.descCol)
                return window.gSortDirection;
            if (x.descCol < y.descCol)
                return -window.gSortDirection;
            return window.prefColSortFunction(x,y);
        }
        window.gSortFunctions.descCol = descColSortFunction;

        // Refresh the tree to show the new column
        tree.treeBoxObject.invalidate();

        function unload(event)
        {
            debug("unload page");

            window.removeEventListener("unload",unload,false);
            xulWindow.removeEventListener("aboutconfigdesc-unload",unload,false);

            prefBranch.removeObserver("",prefListener);

            // Reload about:config to reset it after the extension is shut down
            if (event.type == "aboutconfigdesc-unload")
                window.setTimeout(function() {
                    debug("reload page");
                    window.document.location.reload();
                }, 100);
        }
        window.addEventListener("unload",unload,false);
        xulWindow.addEventListener("aboutconfigdesc-unload",unload,false);
    }
    catch (e) { Components.utils.reportError(e); }
}

function loadDefaultPrefs()
{   
    // Define preference locations
    var prefsGoannaTycho = "resource://gre/goanna.js";
    var prefsGoannaUXP = "resource://gre/greprefs.js";   
    var prefsPlatform = "resource://gre/defaults/pref/";
    var prefsApplication = "resource:///defaults/preferences/";
    
    // Default gre preferences
    var grePrefs = prefsGoannaUXP;
    
    // Get application name
    var appName = Services.appinfo.name;

    // If Pale Moon then override appName and grePrefs
    // When Pale Moon goes UXP and everything is finalized this will have to be revised
    // Though the appname override will have to stay because appinfo.name returns "Pale Moon" but we want the non-spaced version
    if (appName == "Pale Moon" || Services.appinfo.ID == "{8de7fcbb-c55c-4fbe-bfc5-fc555c87dbc4}") {
        grePrefs = prefsGoannaTycho;
        appName = "palemoon";
    }
    else {
        appName = "firefox";
    }
    
    var files =
    [
        prefsApplication + appName + ".js",
        prefsApplication + appName + "-branding.js",
        prefsApplication + appName + "-l10n.js",
        prefsPlatform + "services-sync.js",
        grePrefs
    ];

    var defaultPrefsComments = {};

    for (var i in files)
    {
        var lines = loadDefaultPrefsFile(files[i]);
        if (!lines)
            continue;
        var pref, donePrefComment;
        while (lines.length)
        {
            var line = lines.pop();
            if (line[0] != '/')
            {
                pref = line.split(/"/g)[1];
                if (defaultPrefsComments[pref])  // if there's already a comment recorded for this pref, then it was from a previously checked file; skip this one
                    donePrefComment = true;
                else
                    donePrefComment = false;
                var side = line.split(/\)\s*;\s*/)[1];
                if (side && side[0] == '/')  // if there's a comment to the side, parse it on the next iteration
                    lines.push(side);
                continue;
            }
            else if (line.length == 0 ||               // assume everything not directly next to the pref isn't for it (at least directly)
                     line.substr(0,7) == "//@line" ||  // skip build system notes (?)
                     /pref\s*\(.*\)/.test(line) )      // skip commented out prefs (and their comments)
            {
                donePrefComment = true;
                continue;
            }
            if (!donePrefComment && pref)
            {
                line = line.replace(/(^\/\/|^\/\*|\*\/$)/g,"").trim();  // remove comment syntax ("//", "/*", & "*/")
                line = " " + line;                                      // leave in a leading space so multiline comments show correctly when fitting into one line in the tree
                if (!defaultPrefsComments[pref])
                    defaultPrefsComments[pref] = line;
                else
                    defaultPrefsComments[pref] = line + "\n" + defaultPrefsComments[pref];
            }
        }
    }

    return defaultPrefsComments;
}

function loadDefaultPrefsFile(path)
{
    try
    {
        var uri = Services.io.newURI(path,null,null);
        var channel = Services.io.newChannelFromURI(uri);
        var stream = channel.open();  // Sync open (not done on application startup)
        var scriptableStream = Components.classes["@mozilla.org/scriptableinputstream;1"]
                                         .createInstance(Components.interfaces.nsIScriptableInputStream);
        scriptableStream.init(stream);
        var data = scriptableStream.readBytes(stream.available());
        stream.close();
        if (!data)
            return null;
        debug("loaded default prefs file " + path);
        return data.split('\n');
    }
    catch (e)
    {
        //Components.utils.reportError(e);
        return null;
    }
}

function safeGetBoolPref(pref)  // Returns bool pref value, or false if there's an error or it does not exist
{
    try { return Services.prefs.getBoolPref(pref); }
    catch (e) { return false; }
}

function debug(msg)
{
    if (safeGetBoolPref("extensions.configdescriptions.debug"))
        Services.console.logStringMessage("ConfigDesc: " + msg);
}